# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp._canvas._polygon.polygon import Polygon
from openexp._canvas._element.legacy import LegacyElement
import pygame


class Legacy(LegacyElement, Polygon):

    def prepare(self):

        pygame.draw.polygon(self.surface, self.color.backend_color,
                            [self.to_xy(x, y) for x, y in self.vertices],
                            0 if self.fill else self.penwidth)



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore, QtWidgets
from libqtopensesame.validators._base_validator import BaseValidator


class DurationValidator(BaseValidator):

    r"""A validator for durations."""
    def __init__(self, main_window, default=u'keypress'):

        super().__init__(main_window, default=default)

    def is_valid(self, val):

        if val in [u'keypress', u'mouseclick', u'sound'] \
                or self.experiment.syntax.contains_variables(val):
            return True
        try:
            int(val)
            return True
        except:
            return False



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.oslogging import oslogger
from libopensesame.exceptions import OSException, UnexpectedError
import platform
import sys
import os
import signal
import multiprocessing


class OutputChannel:
    """Passes messages from child process back to main process.
    
    Parameters
    ----------
    channel:
        A multiprocessing.JoinableQueue object that is referenced from the
        main process.
    orig:
        The original stdout or stderr to also print the messages to.
    """
    def __init__(self, channel, orig=None):
        self.channel = channel
        self.orig = orig

    def write(self, m):
        """Writes a message to the queue.

        Parameters
        ----------
        m:
            The message to write. Should be a string or an (Exception,
            traceback) tuple.
        """
        self.channel.put(m)

    def flush(self):
        """Dummy function to mimic the stderr.flush() function."""
        if self.orig:
            self.orig.flush()
        else:
            pass

    def isatty(self):
        """Indicates that the output is not attached to a terminal."""
        return False


class ExperimentProcess(multiprocessing.Process):
    """Creates a new process to run an experiment in.

    Parameters
    ----------
    exp:
        An instance of libopensesame.experiment.experiment
    output:
        A reference to the queue object created in and used to communicate with
        the main process.
    """
    def __init__(self, exp, output):
        multiprocessing.Process.__init__(self)
        self.output = output
        # The experiment object is troublesome to serialize,
        # therefore pull out all relevant data to pass on to the new process
        # and rebuild the exp object in there.
        self.script = exp.to_string()
        self.pool_folder = exp.pool.folder()
        self.subject_nr = exp.var.subject_nr
        self.experiment_path = exp.experiment_path
        self.fullscreen = exp.var.fullscreen == u'yes'
        self.logfile = exp.logfile
        self.killed = False

    def run(self):
        """
        Everything in this function is run in a new process, therefore all
        import statements are put in here. The function reroutes all output to
        stdin and stderr to the pipe to the main process so OpenSesame can
        handle all prints and errors.
        """
        import os
        import sys
        from libopensesame import misc
        from libopensesame.experiment import experiment
        # Under Windows, change the working directory to the OpenSesame folder,
        # so that the new process can find the main script.
        if os.name == u'nt':
            os.chdir(misc.opensesame_folder())
            os.environ['PATH'] += ';' + os.getcwd()
        # Reroute output to OpenSesame main process, so everything will be
        # printed in the Debug window there.
        pipeToMainProcess = OutputChannel(self.output)
        sys.stderr = sys.stdout = pipeToMainProcess
        oslogger.start(u'runtime')
        # First initialize the experiment and catch any resulting Exceptions
        try:
            exp = experiment(
                string=self.script, pool_folder=self.pool_folder,
                experiment_path=self.experiment_path,
                fullscreen=self.fullscreen, subject_nr=self.subject_nr,
                logfile=self.logfile)
        except Exception as e:
            if not isinstance(e, OSException):
                e = UnexpectedError('An unexpected error occurred while '
                                    'building the experiment')
            # Communicate the exception and exit with error
            self.output.put(e)
            sys.exit(1)
        oslogger.info(u'Starting experiment as %s' % self.name)
        # Run the experiment and catch any Exceptions.
        e_run = None
        exp.set_output_channel(self.output)
        try:
            exp.run()
            oslogger.info('experiment finished!')
        except Exception as e:
            if not isinstance(e, OSException):
                e = UnexpectedError('An unexpected error occurred while '
                                    'running the experiment')
            e_run = e
        # Communicate the exception. We do this before calling exp.end()
        # because Python may crash in this step and we need to send the
        # exception before that happens.
        if e_run is not None:
            self.output.put(e_run)
        # End the experiment and catch any Exceptions. These exceptions are just
        # printed out and not explicitly passed on to the user, because they are
        # less important than the run-related exceptions.
        try:
            exp.end()
        except Exception as e_exp:
            oslogger.error(
                u'An Exception occurred during exp.end(): %s' % e_exp)
        exp.transmit_workspace()
        # Exit with error status if an exception occurred.
        if e_run is not None:
            sys.exit(1)
        # Exit with success
        sys.exit(0)

    def kill(self):

        oslogger.info(u'killing experiment process')
        try:
            os.kill(
                self.pid,
                signal.SIGKILL if hasattr(
                    signal, u'SIGKILL') else signal.SIGTERM
            )
        except OSError:
            oslogger.warning(u'failed to kill experiment process')
        self.killed = True



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.syntax import Syntax
import re

FIX_CURLY_BRACES = re.compile(r'\{([a-zA-Z_][a-zA-Z0-9_]*)\}')


class QtSyntax(Syntax):

    def eval_text(self, txt, round_float=False):
        return txt
    
    def fix_conditional_expression(self, cond):
        return FIX_CURLY_BRACES.sub(r'\1', cond)
    

# Alias for backwards compatibility
qtsyntax = QtSyntax



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os.path
import sys
import importlib.util
import qtawesome as qta
from openexp import resources
from libopensesame.oslogging import oslogger
from libqtopensesame.misc.config import cfg
from qtpy import QtGui, QtWidgets, QtCore

available_themes = [u'default', u'monokai']

# Fontawesome 4 (with the fa. prefix) has been removed in recent updates of
# qtawesome. This is used by various packages, so here we monkeypatch the icon()
# function so that it at least doesn't crash, and falls back to a circle icon.
def fix_qta_icon(fnc):
    
    def inner(qta_name, *args, **kwargs):
        if qta_name.startswith('fa.'):
            oslogger.warning('fa. prefix is deprecated, use fa6')
            qta_name = 'fa6.' + qta_name[3:]
        try:
            return fnc(qta_name, *args, **kwargs)
        except Exception:
            oslogger.error(f'{qta_name} is not defined')
            return fnc('fa6.circle', *args, **kwargs)
        
    return inner

qta.icon = fix_qta_icon(qta.icon)


class Theme:

    """Handles the GUI theme"""
    default_icon_size = 32

    def __init__(self, main_window, theme=None):
        """
        Constructor

        Arguments:
        main_window -- the main_window object

        Keyword arguments:
        theme -- the theme to be used or None to use config (default=None)
        """
        self.main_window = main_window
        self.fallback_icon = QtGui.QIcon(resources['theme/fallback.png'])
        self.theme = cfg.theme if theme is None else theme
        self.theme_folder = resources[f'theme/{self.theme}']
        self._icon_theme_path = self.theme_folder
        oslogger.debug(u"theme = '%s' (%s)" % (self.theme, self.theme_folder))
        # The theme folder must exist, and contain a file called __theme__.py,
        # if not, we fall back to the default theme, which is assumed to always
        # exist.
        if self.theme_folder is None or not os.path.exists(
                os.path.join(self.theme_folder, u'__theme__.py')
        ):
            oslogger.warning(
                u"theme '%s' does not exist, using 'default'" % theme,
            )
            self.theme = u"default"
            self._icon_theme_path = self.theme_folder = \
                resources[f'theme/{self.theme}']
        self.theme_info = os.path.join(self.theme_folder, u"__theme__.py")
        if os.path.exists(self.theme_info):
            # Load the theme module from file
            spec = importlib.util.spec_from_file_location(
                self.theme,
                safe_str(self.theme_info, enc=sys.getfilesystemencoding()))
            info = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(info)
            with safe_open(os.path.join(self.theme_folder, info.qss)) as fd:
                self._qss = fd.read()
            self._icon_map = info.icon_map
            self._icon_theme = info.icon_theme
            if hasattr(info, 'qdatamatrix'):
                from qdatamatrix import _qcell
                for key, color in info.qdatamatrix.items():
                    setattr(_qcell, key, color)
            if hasattr(info, 'icon_theme_path'):
                self._icon_theme_path = \
                    resources[f'theme/{info.icon_theme_path}']
        self.load_icon_map()
        self.apply_theme(self.main_window)

    @property
    def experiment(self):
        return self.main_window.experiment

    def apply_theme(self, widget):
        """
        Apply the theme to a QWidget, i.e. load the stylesheet and the icons

        Arguments:
        widget -- a QWidget
        """
        if hasattr(widget, u'setStyleSheet'):
            widget.setStyleSheet(self._qss)
        if hasattr(widget, u"ui"):
            self.load_icons(widget.ui)

    def qfileicon(self, path):
        r"""Gets an filetype icon for a file.

        Parameters
        ----------
        path
            The full path to the file. The file must exist.

        Returns
        -------
        QIcon
        """
        try:
            import fileinspector
        except ImportError:
            return QtWidgets.QFileIconProvider().icon(QtCore.QFileInfo(path))
        filetype = fileinspector.determine_type(path, output='xdg')
        if filetype is None:
            filetype = u'text-x-generic'
        category = fileinspector.determine_category(filetype)
        if category is None:
            fallback = u'text-x-generic'
        else:
            fallback = category + u'-x-generic'
        return QtGui.QIcon.fromTheme(
            filetype,
            QtGui.QIcon.fromTheme(fallback)
        )

    def qicon(self, icon):
        """
        desc:
                Gets an icon from the theme.

        arguments:
                icon:	One of the following:
                                - A QIcon, which returned as is
                                - The name of an image file
                                - The name of a plug-in with a hardcoded icon
                                - The name of an entry in the icon map
                                - A theme icon name.

        returns:
                desc:	An icon, or a fallback icon if the specified wasn't found.
                type:	QIcon
        """
        if isinstance(icon, QtGui.QIcon):
            return icon
        if os.path.isfile(icon):
            qicon = QtGui.QIcon()
            if icon.endswith(u'_large.png'):
                size = 32
            else:
                size = 16
            qicon.addFile(icon, size=QtCore.QSize(size, size))
            return qicon
        if (
                hasattr(self, u'experiment') and
                (
                    u'%s_large.png' % icon in self.experiment.resources or
                    '%s.png' in self.experiment.resources
                )
        ):
            qicon = QtGui.QIcon()
            if u'%s_large.png' % icon in self.experiment.resources:
                qicon.addFile(resources[f'{icon}_large.png'],
                              size=QtCore.QSize(32, 32))
            if u'%s.png' % icon in self.experiment.resources:
                qicon.addFile(resources[f'{icon}.png'],
                              size=QtCore.QSize(16, 16))
            return qicon
        if icon in self.icon_map:
            name = self.icon_map[icon][0]
        else:
            name = icon
        icon = QtGui.QIcon.fromTheme(name, self.fallback_icon)
        if icon.name() != name:
            oslogger.debug(
                u'missing icon {}, using {}'.format(name, icon.name())
            )
        return icon

    def qpixmap(self, icon, size=None):
        """
        Get an icon from the theme

        Arguments:
        icon -- the icon name

        Keyword arguments:
        size -- the size of the icon or None for default (default=None)

        Returns:
        A QPixmap
        """
        if isinstance(icon, str) and os.path.exists(icon):
            return QtGui.QPixmap(icon)
        if size is None:
            if icon in self.icon_map:
                size = self.icon_map[icon][1]
            else:
                size = self.default_icon_size
        size = QtCore.QSize(size, size)
        # Pixmap should never return a pixmap larger than the requested size.
        # However, this does happen on some versions of Qt4. Therefore, we check
        # the size, and scale if necessary.
        pixmap = self.qicon(icon).pixmap(size)
        if pixmap.size() != size:
            pixmap = pixmap.scaled(size)
        return pixmap

    def qlabel(self, icon, size=None):
        """
        Get an icon from the theme

        Arguments:
        icon -- the icon name

        Keyword arguments:
        size -- the size of the icon or None for default (default=None)

        Returns:
        A QLabel
        """
        label = QtWidgets.QLabel()
        label.setPixmap(self.qpixmap(icon, size=size))
        return label

    def load_icon_map(self):
        """Load the icon map"""
        self.original_theme = QtGui.QIcon.themeName()
        if os.path.exists(
                os.path.join(self._icon_theme_path, self._icon_theme)
        ):
            oslogger.debug(u"using custom icon theme")
            QtGui.QIcon.setThemeSearchPaths(
                QtGui.QIcon.themeSearchPaths() + [self._icon_theme_path]
            )
            QtGui.QIcon.setThemeName(self._icon_theme)
        else:
            oslogger.warning(u"using default icon theme, icons may be missing")
        self.icon_map = {}
        path = os.path.join(self.theme_folder, self._icon_map)
        with safe_open(path) as fd:
            for line in fd:
                line = line.split(u',')
                if len(line) != 3:
                    continue
                try:
                    size = int(line[2])
                except ValueError:
                    size = 32
                alias = line[0].strip()
                name = line[1].strip()
                if alias in self.icon_map:
                    oslogger.debug(
                        u"alias '%s' already in icon map, overwriting" % alias,
                        reason=u"warning"
                    )
                self.icon_map[alias] = name, size

    def load_icons(self, ui):
        """
        Add icons to all icon-supporting widgets in a ui object. Icons can
        be specified through a `qta:[icon name]` text on the widget, or through
        the icon_map, where the object name is the look-up key.

        Parameters
        ----------
        ui
            The ui object to load icons into
        """
        if ui is None:
            return
        for key, value in ui.__dict__.items():
            if not isinstance(value, QtCore.QObject):
                continue
            if hasattr(value, 'text') and value.text().startswith('qta:'):
                try:
                    value.setIcon(qta.icon(value.text()[4:]))
                    value.setText('')
                except Exception as e:
                    oslogger.warning(f'failed to load qtawesome icon: {e}')
                    
            elif key in self.icon_map:
                if hasattr(value, 'setIcon'):
                    value.setIcon(self.qicon(key))
                elif hasattr(value, 'setPixmap'):
                    value.setPixmap(self.qpixmap(key))

    def set_toolbar_size(self, size):
        """
        Control the size of the icons in the toolbar

        Arguments:
        size -- a size in pixels
        """
        self.main_window.ui.toolbar_main.setIconSize(QtCore.QSize(size, size))
        self.main_window.ui.toolbar_items.setIconSize(QtCore.QSize(size, size))
        self.main_window.ui.toolbar_items.build()

    def resource(self, fname):
        """
        Retrieves the path to a resource within the theme folder.

        Arguments:
        fname	--	The resource filename.

        Returns:
        The full path to the resource file in the theme folder.
        """
        return os.path.join(self.theme_folder, fname)


# Alias for backwards compatibility
theme = Theme
